#!/usr/bin/env python
"""
  The module mftp is a numerical solver for computing the maximum flux
  transition path Z(s), 0 < s < 1:
  D^{-1} Z_s // -beta F_zeta + (D^{-1}Z_s)_s/(Z_s^T D^{-1} Z_s).
  In this module we implemented the semi-implicit simplified string method
  described in the arxiv paper. 
  The module consists of four functions: refine_path calling mftp, and mftp
  calling thomas_solver and normalize.

  An example is provided to demonstrate the algorithm at the end.
  The problem is to find MFTP for the three-hole potential in the arxiv paper.
"""
import numpy as np
import time
import numpy.linalg as la

try:
    from mpi4py import MPI
except:
    try:
        import FMPI as MPI  # FAKE mpi4py
    except:
        print "MPI4PY NOT FOUND"

# The function 'refine_path' takes three arguments, msys (molecular system),
# path, and misc (parameters):
#   * path (no_img x dim_cv array) is an initial path
#   * misc (a dictionary) has keys:
#       - TOL (real), a stopping criterion, also returns as an error term
#       - noiter (integer), a stopping criterion, also returns as a real computational effort
#       - kcst (real), spring const of harmonic restrains
#   * msys should have two attributes: the system temperature
#     and a method called mean_force;
#     In particular, mean_force returns free energy gradient F (dim_cv x 1 array)
#     and proto-diffusion tensor matrix D (dimcv x dim_cv array) with harmonic
#     restrains centered at path[j] (dim_cv x 1 array) and spring constant
#     misc['kcst']
# The output of the function are
#   * pmf (potential of mean force), a no_img x 1 array 
#   * committor, a no_img x 1 array
#   * flow_rate (flow rate), a no_img x 1 array
#   * path (maximum flux transition path), a no_img x dim_cv array
def refine_path(msys, path, misc):   
    (norepl, nocv) = path.shape
    beta = 1.67*300/msys.temp
    # MPI parallel code starts here
    noproc = MPI.COMM_WORLD.Get_size()
    rank = MPI.COMM_WORLD.Get_rank()
    noloop = (norepl-1)/noproc +1
    cycle = 0
    dist = 100.0
    while (dist > misc['TOL'] and cycle < misc['noiter']):
        start = time.time()
        allD = np.zeros((noproc,noloop,nocv,nocv),'float')
        allF = np.zeros((noproc,noloop,nocv),'float')
        for idrepl in range(rank, norepl, noproc):
            j = idrepl/noproc
            [allF[rank][j],allD[rank][j]] = msys.mean_force(path[idrepl],idrepl,
                                                            misc['kcst'])
        allD = MPI.COMM_WORLD.reduce(allD)
        allF = MPI.COMM_WORLD.reduce(allF)
        if rank == 0:
            D = np.zeros((norepl,nocv,nocv),'float')                                
            F = np.zeros((norepl,nocv),'float')
            for j in range(norepl):
                D[j] = allD[j%noproc][j/noproc]
                F[j] = allF[j%noproc][j/noproc]
            path, dist = mftp(D, F, path, beta, misc['tau'])
        cycle = cycle + 1
        if rank == 0:
            print 'cycle =', cycle, '\t dist =', dist, \
                  '\t time =', time.time()-start
        path = MPI.COMM_WORLD.bcast(path,root=0)
        dist = MPI.COMM_WORLD.bcast(dist,root=0)
##    The function extra will be added later.
##    for idrepl in range(rank, norepl, noproc):
##        j = idrepl/noproc
##        [allD[rank][j],allF[rank][j]] = msys.mean_force(path,idrepl)
##    allD = MPI.COMM_WORLD.reduce(allD)
##    allF = MPI.COMM_WORLD.reduce(allF)
##    for j in range(norepl):
##        D[j] = allD[j%noproc][j/noproc]
##        F[j] = allF[j%noproc][j/noproc]
##    if myrank == 0:
##        pmf, committor, flow_rate = extras(D, F, Z)
    misc['TOL'] = dist
    misc['noiter'] = cycle
    return [path]

# The function mftp updates a path following semi-implicit simplified string
# method (downhill iterating + normalization). 
# The input are
#    * discretized path Z (a no_img x dim_cv array)
#    * proto-diffusion tensor matrxi D at Z (a no_img x dim_cv x dim_cv array)
#    * free energy gradient Fzeta at Z (a no_img x dim_cv array)
#    * inverse temperature beta (real)
#    * step size tau (real)
# The output are
#    * new path after iteration newZ (a no_img x dim_cv array)
#    * the distance between Z and newZ (real)
def mftp(D, Fzeta, Z, beta, tau):
    # compute the right hand side
    (norepl,nocv) = Z.shape
    inv_c = np.zeros((norepl), 'float')           
    RH = np.zeros((norepl,nocv),'float')
    for j in range(1,norepl-1):
        Zplus = Z[j+1]-Z[j]
        Zminus = Z[j]-Z[j-1]
        D_inv = la.inv(D[j])
        DZplus = np.dot(D_inv, Zplus)
        DZminus = np.dot(D_inv, Zminus)
        inv_c[j] = 2.0/(np.dot(Zplus, DZplus) + np.dot(Zminus, DZminus))
        RH[j] = Z[j] - tau*(beta*np.dot(D[j], Fzeta[j]) +
                            (inv_c[j]/2.0)*(np.dot(D[j+1] - D[j], DZplus) +
                                            np.dot(D[j] - D[j-1], DZminus)))
    RH[0] = Z[0] - tau*beta*np.dot(D[0], Fzeta[0])
    RH[norepl-1] = Z[norepl-1] - tau*beta*np.dot(D[norepl-1], Fzeta[norepl-1]) 
    # change the order of variables to transform block tridiagonal to block diagonal matrix
    RH = np.transpose(RH)
    newZ = np.zeros((nocv,norepl),'float')
    # compute the left hand side: the mass matrix 
    a = -tau*inv_c
    b = 1 + 2*tau*inv_c
    c = -tau*inv_c
    # solve equation
    for j in range(nocv):
        newZ[j] = thomas_solver(a,b,c,RH[j],norepl)
    # switch back to the original order of variables
    newZ = np.transpose(newZ)
    # normalize
    newZ = normalize(newZ)
    dist = np.max([la.norm(newZ[j] - Z[j]) for j in range(norepl)])
    return newZ, dist


# The function 'normalize' is normalization of eqal arc length.
# Input are a (no_img x dim_cv) array, and output is also a (no_img x dim_cv) array
def normalize(path):
    (norepl, nocv) = path.shape
    L = np.linspace(0.0, 0.0, norepl)
    for i in range(1,norepl): 
        diff = path[i] - path[i-1]
        L[i] = L[i-1] + np.sqrt(np.dot(diff,diff))

    newpath = np.zeros((norepl,nocv),'float')
    newpath[0] = path[0]
    newpath[norepl-1] = path[norepl-1]
    for i in range(1,norepl-1):
        s = (i/(float(norepl-1)))*L[norepl-1]
        j=0
        while s>=L[j]:
            j += 1
        diff = path[j] - path[j-1]
        newpath[i] = path[j-1]+(s-L[j-1])/np.sqrt(np.dot(diff,diff))*diff
    return newpath


# Thomas (tridiagonal matrix) algorithm is implented to solve
# a_i x_{i-1} + b_i x_i + c_i x_{i+1} = d_i, where a_0 = 0 and c_n = 0
def thomas_solver(a,b,c,d,n):
    # forward substitution
    cp = np.linspace(0.0, 0.0, n)
    cp[0] = c[0]/b[0]
    for j in range(1,n-1):
        cp[j] = c[j]/(b[j] - cp[j-1]*a[j])
    dp = np.linspace(0.0, 0.0, n)
    dp[0] = d[0]/b[0]
    for j in range(1,n):
        dp[j] = (d[j]-dp[j-1]*a[j])/(b[j]-cp[j-1]*a[j])
    #back substitution
    x = np.linspace(0.0, 0.0, n)
    x[n-1]=dp[n-1]
    for j in range(n-2,-1,-1):
        x[j]=dp[j]-cp[j]*x[j+1]
    return x


if __name__ == '__main__':

    def example_energy():
        # number of partition points and contour window.
        # interface could be nopart, and window [-3.0,3.0,-2.5,3.5]
        nopart = 100  # number of mesh points in x-direction
        XB = -3.0     # starting coordinate in x-direction
        XE = 3.0      # ending coordinate in x-direction
        YB = -3.0     # starting coordinate in y-direction
        YE = 3.0      # ending coordinate in y-direction
        epsilon = 0.001
        [x,y] = np.meshgrid(np.linspace(XB,XE,nopart),np.linspace(YB,YE,nopart));
        aa = np.array([-4, -1, -1, -1])
        bb = np.array([0, 0, 0, 0])
        cc = np.array([-1, -1, -1, -1])
        AA = np.array([-4, -5, -5, 8])
        XX = np.array([0, 1, -1, 0])
        YY = np.array([2.75, 0.15, 0, -0.5])
        # calculation of the potential on grid points
        V = AA[0]*np.exp(aa[0]*pow(x-XX[0],2)+bb[0]*(x-XX[0])*(y-YY[0])+cc[0]*pow(y-YY[0],2));
        for j in range(1,4):
            V =  V + AA[j]*np.exp(aa[j]*pow(x-XX[j],2)+bb[j]*(x-XX[j])*(y-YY[j])+cc[j]*pow(y-YY[j],2))
        V = V + epsilon*(np.power(x,4)+np.power(y,4))
        return [x,y,V]

    def example_force(x,y):
        # parameters in potential energy function
        epsilon = 0.001
        aa = np.array([-4, -1, -1, -1])
        bb = np.array([0, 0, 0, 0])
        cc = np.array([-1, -1, -1, -1])
        AA = np.array([-4, -5, -5, 8])
        XX = np.array([0, 1, -1, 0])
        YY = np.array([2.75, 0.15, 0, -0.5])

        ee = AA[0]*np.exp(aa[0]*pow(x-XX[0], 2) +
                          bb[0]*(x-XX[0])*(y-YY[0]) + cc[0]*pow(y-YY[0], 2))
        dVx = (2*aa[0]*(x-XX[0]) + bb[0]*(y-YY[0]))*ee
        dVy = (bb[0]*(x-XX[0]) + 2*cc[0]*(y-YY[0]))*ee
        for j in range(1,4):
            ee = AA[j]*np.exp(aa[j]*pow(x-XX[j],2) +
                              bb[j]*(x-XX[j])*(y-YY[j]) + cc[j]*pow(y-YY[j],2))
            dVx = dVx + (2*aa[j]*(x-XX[j]) + bb[j]*(y-YY[j]))*ee
            dVy = dVy + (bb[j]*(x-XX[j]) + 2*cc[j]*(y-YY[j]))*ee
        dVx = dVx + epsilon*4*pow(x,3)
        dVy = dVy + epsilon*4*pow(y,3)
        return np.array([dVx,dVy])
    
    ###########################################################################
    # main program starts here
    class MSys:
        def __init__(self):
            self.temp = 300.0  
            self.min = []
            self.equil = []
            self.sampling = []
            self.colvar = [' x ', ' y ']
            self.mass = [1.0, 1.0]
        # The key method 'mean_force' has to be defined, providing mean force F and
        # proto-diffusion tensor matrix D
        def mean_force(self, zeta, j, kcst):
            F = example_force(zeta[0], zeta[1])
            beta = 1.67*300.0/self.temp
            D = (1.0/(2.0*beta))*np.identity((2), 'float')
            return F, D
    # initialize object of molecular system
    msys = MSys()
    # initialize a path
    initpath = np.array([[-1.0, 0.0], [-0.9, 0.0], [-0.8, 0.0], [-0.7, 0.0],
                         [-0.6, 0.0], [-0.5, 0.0], [-0.4, 0.0], [-0.3, 0.0],
                         [-0.2, 0.0], [-0.1, 0.0], [0.0, 0.0], [0.1, 0.0],
                         [0.2, 0.0], [0.3, 0.0], [0.4, 0.0], [0.5, 0.0],
                         [0.6, 0.0], [0.7, 0.0], [0.8, 0.0], [0.8, 0.0],
                         [1.0, 0.0]])
    # Control parameters
    misc = {'TOL':0.000001, 'noiter':100, 'kcst': 2000, 'tau':0.1}

    # refine path 
    [tranpath] = refine_path(msys, initpath, misc)

    # print out initial path and transition path and save path into tranpath.dat
    if MPI.COMM_WORLD.Get_rank() == 0:
        np.savetxt('tranpath.dat', tranpath)
        try:
            import pylab as p 
            [x,y,V] = example_energy()
            p.figure()
            p.hold(True)
            p.contour(x, y, V, 40)
            p.plot(initpath[:,0], initpath[:,1], 'ro-')
            p.plot(tranpath[:,0], tranpath[:,1], 'kd-')
            p.hold(False)
            p.show()
        except:
            print 'Pylab is not installed. The transition path is saved.'